from unittest import TestCase
import numpy as np
import os.path
from redisai import Client, DType, Backend, Device, Tensor, BlobTensor
from ml2rt import load_model
from redis.exceptions import ResponseError


MODEL_DIR = os.path.dirname(os.path.abspath(__file__)) + '/testdata'


class TensorTestCase(TestCase):
    def testTensorShapes(self):
        t = Tensor(DType.float, [4], [1, 2, 3, 4])
        self.assertEqual([4], t.shape)
        t = BlobTensor.from_numpy(np.array([[1, 2, 3], [4, 5, 6]]))
        self.assertEqual([2, 3], t.shape)


class ClientTestCase(TestCase):
    def setUp(self):
        super(ClientTestCase, self).setUp()
        self.get_client().flushall()

    def get_client(self):
        return Client()

    def test_set_tensor(self):
        con = self.get_client()
        con.tensorset('x', (2, 3), dtype=DType.float)
        values = con.tensorget('x', as_type=Tensor)
        self.assertEqual([2, 3], values.value)

        con.tensorset('x', Tensor.scalar(DType.int32, 2, 3))
        values = con.tensorget('x', as_type=Tensor).value
        self.assertEqual([2, 3], values)
        meta = con.tensorget('x', meta_only=True)
        self.assertTrue('<Tensor(shape=[2] type=DType.int32) at ' in repr(meta))

        self.assertRaises(Exception, con.tensorset, 1)
        self.assertRaises(Exception, con.tensorset, 'x')

    def test_numpy_tensor(self):
        con = self.get_client()
        input_array = np.array([2, 3])
        con.tensorset('x', input_array)
        values1 = con.tensorget('x')
        self.assertTrue(np.allclose([2, 3], values1))
        self.assertEqual(values1.dtype, np.int64)
        self.assertEqual(values1.shape, (2,))
        self.assertTrue((np.allclose(values1, input_array)))

        values2 = con.tensorget('x', as_type=BlobTensor)
        self.assertTrue(np.allclose(input_array, values2.to_numpy()))
        self.assertTrue(np.allclose(values1, values2.to_numpy()))
        ret = con.tensorset('x', values2)
        self.assertEqual(ret, b'OK')

    def test_run_tf_model(self):
        model_path = os.path.join(MODEL_DIR, 'graph.pb')
        bad_model_path = os.path.join(MODEL_DIR, 'pt-minimal.pt')

        model_pb = load_model(model_path)
        wrong_model_pb = load_model(bad_model_path)

        con = self.get_client()
        con.modelset('m', Backend.tf, Device.cpu, model_pb,
                     inputs=['a', 'b'], outputs='mul')

        # wrong model
        self.assertRaises(ResponseError,
                          con.modelset, 'm', Backend.tf, Device.cpu,
                          wrong_model_pb,
                          inputs=['a', 'b'], outputs='mul')
        # missing inputs/outputs
        self.assertRaises(ValueError,
                          con.modelset, 'm', Backend.tf, Device.cpu,
                          wrong_model_pb)

        # wrong backend
        self.assertRaises(ResponseError,
                          con.modelset, 'm', Backend.torch, Device.cpu,
                          model_pb,
                          inputs=['a', 'b'], outputs='mul')

        con.tensorset('a', Tensor.scalar(DType.float, 2, 3))
        con.tensorset('b', Tensor.scalar(DType.float, 2, 3))
        con.modelrun('m', ['a', 'b'], 'c')
        tensor = con.tensorget('c')
        self.assertTrue(np.allclose([4, 9], tensor))
        model_det = con.modelget('m')
        self.assertTrue(model_det['backend'] == Backend.tf)
        self.assertTrue(model_det['device'] == Device.cpu)
        con.modeldel('m')
        self.assertRaises(ResponseError, con.modelget, 'm')

    def test_scripts(self):
        con = self.get_client()
        self.assertRaises(ResponseError, con.scriptset,
                          'ket', Device.cpu, 'return 1')
        script = r"""
def bar(a, b):
    return a + b
"""
        con.scriptset('ket', Device.cpu, script)
        con.tensorset('a', Tensor.scalar(DType.float, 2, 3))
        con.tensorset('b', Tensor.scalar(DType.float, 2, 3))
        # try with bad arguments:
        self.assertRaises(ResponseError,
                          con.scriptrun, 'ket', 'bar', inputs='a', outputs='c')
        con.scriptrun('ket', 'bar', inputs=['a', 'b'], outputs='c')
        tensor = con.tensorget('c', as_type=Tensor)
        self.assertEqual([4, 6], tensor.value)
        script_det = con.scriptget('ket')
        self.assertTrue(script_det['device'] == Device.cpu)
        self.assertTrue(script_det['script'] == script)
        self.assertTrue("def bar(a, b):" in script_det['script'])
        con.scriptdel('ket')
        self.assertRaises(ResponseError, con.scriptget, 'ket')

    def test_run_onnxml_model(self):
        mlmodel_path = os.path.join(MODEL_DIR, 'boston.onnx')
        onnxml_model = load_model(mlmodel_path)
        con = self.get_client()
        con.modelset("onnx_model", Backend.onnx, Device.cpu, onnxml_model)
        tensor = BlobTensor.from_numpy(np.ones((1, 13), dtype=np.float32))
        con.tensorset("input", tensor)
        con.modelrun("onnx_model", ["input"], ["output"])
        outtensor = con.tensorget("output", as_type=Tensor)
        self.assertEqual(int(outtensor.value[0]), 24)

    def test_run_onnxdl_model(self):
        # A PyTorch model that finds the square
        dlmodel_path = os.path.join(MODEL_DIR, 'findsquare.onnx')
        onnxdl_model = load_model(dlmodel_path)
        con = self.get_client()
        con.modelset("onnx_model", Backend.onnx, Device.cpu, onnxdl_model)
        tensor = BlobTensor.from_numpy(np.array((2, 3), dtype=np.float32))
        con.tensorset("input", tensor)
        con.modelrun("onnx_model", ["input"], ["output"])
        outtensor = con.tensorget("output")
        self.assertTrue(np.allclose(outtensor, [4.0, 9.0]))


# TODO: image/blob tests; more numpy tests..



from redis import StrictRedis
from typing import Union, Any, AnyStr, ByteString, Collection, Type

try:
    import numpy as np
except ImportError:
    np = None

from .constants import Backend, Device, DType
from .utils import str_or_strsequence, to_string
from .tensor import Tensor, BlobTensor


class Client(StrictRedis):
    """
    RedisAI client that can call Redis with RedisAI specific commands
    """
    def loadbackend(self, identifier: AnyStr, path: AnyStr) -> AnyStr:
        """
        RedisAI by default won't load any backends. User can either explicitly
        load the backend by using this function or let RedisAI load the required
        backend from the default path on-demand.

        :param identifier: String representing which backend. Allowed values - TF, TORCH & ONNX
        :param path: Path to the shared object of the backend
        :return: byte string represents success or failure
        """
        return self.execute_command('AI.CONFIG LOADBACKEND', identifier, path)

    def modelset(self,
                 name: AnyStr,
                 backend: Backend,
                 device: Device,
                 data: ByteString,
                 inputs: Union[AnyStr, Collection[AnyStr], None] = None,
                 outputs: Union[AnyStr, Collection[AnyStr], None] = None
                 ) -> AnyStr:
        args = ['AI.MODELSET', name, backend.value, device.value]
        if backend == Backend.tf:
            if not(all((inputs, outputs))):
                raise ValueError(
                    'Require keyword arguments input and output for TF models')
            args += ['INPUTS'] + str_or_strsequence(inputs)
            args += ['OUTPUTS'] + str_or_strsequence(outputs)
        args += [data]
        return self.execute_command(*args)

    def modelget(self, name: AnyStr) -> dict:
        rv = self.execute_command('AI.MODELGET', name)
        return {
            'backend': Backend(to_string(rv[0])),
            'device': Device(to_string(rv[1])),
            'data': rv[2]
        }

    def modeldel(self, name: AnyStr) -> AnyStr:
        return self.execute_command('AI.MODELDEL', name)

    def modelrun(self,
                 name: AnyStr,
                 inputs: Union[AnyStr, Collection[AnyStr]],
                 outputs: Union[AnyStr, Collection[AnyStr]]
                 ) -> AnyStr:
        args = ['AI.MODELRUN', name]
        args += ['INPUTS'] + str_or_strsequence(inputs)
        args += ['OUTPUTS'] + str_or_strsequence(outputs)
        return self.execute_command(*args)

    def tensorset(self,
                  key: AnyStr,
                  tensor: Union[Tensor, np.ndarray, list, tuple],
                  shape: Union[Collection[int], None] = None,
                  dtype: Union[DType, None] = None) -> Any:
        """
        Set the values of the tensor on the server using the provided Tensor object
        :param key: The name of the tensor
        :param tensor: a `Tensor` object
        :param shape: Shape of the tensor
        :param dtype: data type of the tensor. Required if input is a sequence of ints/floats
        """
        # TODO: tensorset will not accept BlobTensor or Tensor object in the future.
        # Keeping it in the current version for compatibility with the example repo
        if np and isinstance(tensor, np.ndarray):
            tensor = BlobTensor.from_numpy(tensor)
        elif isinstance(tensor, (list, tuple)):
            if shape is None:
                shape = (len(tensor),)
            tensor = Tensor(dtype, shape, tensor)
        args = ['AI.TENSORSET', key, tensor.type.value]
        args += tensor.shape
        args += [tensor.ARGNAME]
        args += tensor.value
        return self.execute_command(*args)

    def tensorget(self,
                  key: AnyStr, as_type: Type[Tensor] = None,
                  meta_only: bool = False) -> Union[Tensor, BlobTensor]:
        """
        Retrieve the value of a tensor from the server. By default it returns the numpy array
        but it can be controlled using `as_type` argument and `meta_only` argument.
        :param key: the name of the tensor
        :param as_type: the resultant tensor type. Returns numpy array if None
        :param meta_only: if true, then the value is not retrieved,
            only the shape and the type
        :return: an instance of as_type
        """
        # TODO; We might remove Tensor & BlobTensor in the future and `tensorget` will return
        # python list or numpy arrays or a namedtuple
        if meta_only:
            argname = 'META'
        elif as_type is None:
            argname = BlobTensor.ARGNAME
        else:
            argname = as_type.ARGNAME

        res = self.execute_command('AI.TENSORGET', key, argname)
        dtype, shape = to_string(res[0]), res[1]
        dt = DType.__members__[dtype.lower()]
        if meta_only:
            return Tensor(dt, shape, [])
        elif as_type is None:
            return BlobTensor.from_resp(dt, shape, res[2]).to_numpy()
        else:
            return as_type.from_resp(dt, shape, res[2])

    def scriptset(self, name: AnyStr, device: Device, script: AnyStr) -> AnyStr:
        return self.execute_command('AI.SCRIPTSET', name, device.value, script)

    def scriptget(self, name: AnyStr) -> dict:
        r = self.execute_command('AI.SCRIPTGET', name)
        device = Device(to_string(r[0]))
        return {
            'device': device,
            'script': to_string(r[1])
        }

    def scriptdel(self, name):
        return self.execute_command('AI.SCRIPTDEL', name)

    def scriptrun(self,
                  name: AnyStr,
                  function: AnyStr,
                  inputs: Union[AnyStr, Collection[AnyStr]],
                  outputs: Union[AnyStr, Collection[AnyStr]]
                  ) -> AnyStr:
        args = ['AI.SCRIPTRUN', name, function, 'INPUTS']
        args += str_or_strsequence(inputs)
        args += ['OUTPUTS']
        args += str_or_strsequence(outputs)
        return self.execute_command(*args)



from enum import Enum


class Device(Enum):
    cpu = 'CPU'
    gpu = 'GPU'


class Backend(Enum):
    tf = 'TF'
    torch = 'TORCH'
    onnx = 'ONNX'


class DType(Enum):
    float = 'FLOAT'
    double = 'DOUBLE'
    int8 = 'INT8'
    int16 = 'INT16'
    int32 = 'INT32'
    int64 = 'INT64'
    uint8 = 'UINT8'
    uint16 = 'UINT16'
    uint32 = 'UINT32'
    uint64 = 'UINT64'

    # aliases
    float32 = 'FLOAT'
    float64 = 'DOUBLE'



from typing import Union, ByteString, Collection
import warnings
from .utils import convert_to_num
from .constants import DType
try:
    import numpy as np
except (ImportError, ModuleNotFoundError):
    np = None


class Tensor(object):
    ARGNAME = 'VALUES'

    def __init__(self,
                 dtype: DType,
                 shape: Collection[int],
                 value):
        warnings.warn("Tensor APIs are depricated and "
                      "will be removed from the future release.", UserWarning)
        """
        Declare a tensor suitable for passing to tensorset
        :param dtype: The type the values should be stored as.
            This can be one of Tensor.FLOAT, tensor.DOUBLE, etc.
        :param shape: An array describing the shape of the tensor. For an
            image 250x250 with three channels, this would be [250, 250, 3]
        :param value: The value for the tensor. Can be an array.
            The contents must coordinate with the shape, meaning that the
            overall length needs to be the product of all figures in the
            shape. There is no verification to ensure that each dimension
            is correct. Your application must ensure that the ordering
            is always consistent.
        """
        self.type = dtype
        self.shape = list(shape)
        self.value = value
        if not isinstance(value, (list, tuple)):
            self.value = [value]

    def __repr__(self):
        return '<{c.__class__.__name__}(shape={s} type={t}) at 0x{id:x}>'.format(
            c=self,
            s=self.shape,
            t=self.type,
            id=id(self))

    @classmethod
    def from_resp(cls, dtype: DType, shape: Collection[int], value) -> 'Tensor':
        convert_to_num(dtype, value)
        return cls(dtype, shape, value)

    @classmethod
    def scalar(cls, dtype: DType, *items) -> 'Tensor':
        """
        Create a tensor from a list of numbers
        :param dtype: Type to use for storage
        :param items: One or more items
        :return: Tensor
        """
        return cls(dtype, [len(items)], items)


class BlobTensor(Tensor):
    ARGNAME = 'BLOB'

    def __init__(self,
                 dtype: DType,
                 shape: Collection[int],
                 *blobs: Union['BlobTensor', ByteString]
                 ):
        """
        Create a tensor from a binary blob
        :param dtype: The datatype, one of Tensor.FLOAT, Tensor.DOUBLE, etc.
        :param shape: An array
        :param blobs: One or more blobs to assign to the tensor.
        """
        if len(blobs) > 1:
            blobarr = bytearray()
            for b in blobs:
                if isinstance(b, BlobTensor):
                    b = b.value[0]
                blobarr += b
            size = len(blobs)
            ret_blobs = bytes(blobarr)
            shape = [size] + list(shape)
        else:
            ret_blobs = bytes(blobs[0])

        super(BlobTensor, self).__init__(dtype, shape, ret_blobs)

    @classmethod
    def from_numpy(cls, *nparrs) -> 'BlobTensor':
        blobs = []
        for arr in nparrs:
            blobs.append(arr.data)
        dt = DType.__members__[str(nparrs[0].dtype)]
        return cls(dt, nparrs[0].shape, *blobs)

    @property
    def blob(self):
        return self.value[0]

    def to_numpy(self) -> np.array:
        a = np.frombuffer(self.value[0], dtype=self._to_numpy_type(self.type))
        return a.reshape(self.shape)

    @staticmethod
    def _to_numpy_type(t):
        if isinstance(t, DType):
            t = t.value
        mm = {
            'FLOAT': 'float32',
            'DOUBLE': 'float64'
        }
        if t in mm:
            return mm[t]
        return t.lower()

    @classmethod
    def from_resp(cls, dtype, shape, value) -> 'BlobTensor':
        return cls(dtype, shape, value)



import six
from .constants import DType


def to_string(s):
    if isinstance(s, six.string_types):
        return s
    elif isinstance(s, bytes):
        return s.decode('utf-8')
    else:
        return s  # Not a string we care about


def str_or_strsequence(v):
    if not isinstance(v, (list, tuple)):
        if isinstance(v, six.string_types):
            return [v]
        else:
            raise TypeError('Argument must be a string, list or a tuple')
    return v


def convert_to_num(dt: DType, arr) -> None:
    """
    Recurse value, replacing each element of b'' with the appropriate element.
    Function doesn't return anything but does inplace operation which updates `arr`

    :param dt: Type of tensor | array
    :param arr: The array with b'' numbers or recursive array of b''
    """
    for ix in six.moves.range(len(arr)):
        obj = arr[ix]
        if isinstance(obj, list):
            convert_to_num(dt, obj)
        else:
            if dt in (DType.float, DType.double):
                arr[ix] = float(obj)
            else:
                arr[ix] = int(obj)
